! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: eboxdetect_mod
!
! Authors:  Hermann Brunner - AIP
!          ! Description: Sliding box/matched filter EPIC source searching
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module eboxdetect_mod

  use param  
  use dal
  use eboxdata
  use ebox_io
  use edetect_io_box
  use edetect_gen
   
  implicit none 
  
  private 

  public eboxdetect

contains 

  !---------------------------------------------------------------------
  subroutine eboxdetect
  !---------------------------------------------------------------------

!  author : grh        date: 27-mar-1987 08:56
!  update : cri        date: 8-feb-1988 15:39
!  update : ptk        date: 2-may-1988 16:57
!  update : ptk        date: 17-nov-1992 15:29 
!  update : exsas      date: 14-dec-1992 23:13

!  mdetec              sr      source detection algorithm
!  exdmin              sr      read input files

!  bkg         background count
!  isnr        signal to noise ratio
!  iwind       detection window size
!  nmerk       total number of detected sources
!  snrm        minimum signal to noise ratio
!  src         source count
!  xxms,yyms   source x and y positions

    integer ::    ll,ierr,istat,nmerk,i
    integer, dimension(max_sou) :: iwind, cflag
    real ::       snrm,bkscal
    real, dimension(max_sou) :: xxms,yyms,xerr,yerr
    real, dimension(:,:), allocatable :: src,esrc,bkg,rat,erat,expos, &
                                         isnr,raw_cts,raw_bkg,eef 
    integer, dimension(:,:), allocatable :: maskzero

    character*12  mis_id,det_id

    ierr=0

    call dmpari ! (dmpar,readext,istat)

    i = index(dmexpd,',')
    mis_id = dmexpd(1:i-1)
    det_id = dmexpd(i+1:)

      ll = 0

      call message(AppMsg,Verbose,"Reading input data ...")
      call rd_data


      allocate (src(nim,max_sou))
      allocate (esrc(nim,max_sou))
      allocate (bkg(nim,max_sou))
      allocate (expos(nim,max_sou))
      allocate (rat(nim,max_sou))
      allocate (erat(nim,max_sou))
      allocate (isnr(nim,max_sou))
      allocate (raw_cts(nim,max_sou))
      allocate (raw_bkg(nim,max_sou))
      allocate (eef(nim,max_sou))
      allocate (maskzero(nim,max_sou))

      maskzero=0


      ipa_x = npix(1)
      ipa_y = npix(2)
      idm_x = ipa_x
      idm_y = ipa_y

      nmerk  = 0
      bkscal = 1./bscal
      snrm   = dm_snrm

! source detection algorithm

      call mdetec (idm_x,idm_y,snrm,dm_nrun,ipa_run,dm_larx, &
                   dm_lary,dm_lmap,dm_lloc,bkscal,nmerk,xxms,yyms, &
                   xerr,yerr,isnr,iwind,src,esrc,bkg,expos,rat,erat, &
                   raw_cts,raw_bkg,eef,maskzero,istat)

!     dm_nsrc(iech) = nmerk
!     if(dm_nsrc(iech).gt.0)  found=.true.

      call cluster_check (nmerk, max_sou, xxms, yyms, src, 5.0, cflag)
! writes the results

      call message(AppMsg,Verbose,"Writing source list ...")

      

      call wr_stab(ll,nmerk,xxms(1:nmerk),yyms(1:nmerk),xerr(1:nmerk),yerr(1:nmerk), &
                   src(:,1:nmerk),esrc(:,1:nmerk),bkg(:,1:nmerk),isnr(:,1:nmerk),iwind(1:nmerk), &
                   raw_cts(:,1:nmerk),raw_bkg(:,1:nmerk),eef(:,1:nmerk),maskzero(:,1:nmerk)) 

! close output table
!     ll = -ll
     call boxout(0,-1,ll,nmerk,0,0,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0.,0., &
                 0.,0d0,0d0,0.,0d0,0d0,0.,0.,0.,0.,0.,0.,0.,0,0.,0.,0.)

     call message(AppMsg,Verbose,"Done")

  end subroutine eboxdetect

  !---------------------------------------------------------------------
  subroutine cmass (nx,ny,ifeld,ndimx,ndimy,nsize,xms,yms,xerr,yerr)
  !---------------------------------------------------------------------

!  calculates center of mass of a region of interest in the field

!  update : gih        date: 16-jan-1987 17:03
!  update : cri        date: 19-jan-1988 15:02
!  update : ptk        date: 16-feb-1990 16:36
!  update : exsas      date: 03-aug-1993 17:01

!  a region of interest within the array ifeld is defined by the co-
!  ordinates nx,ny and the square half size nsize. the center of mass
!  of the region of interest is calculated an handed over to xms,yms

!  ifeld               r4  i   input array
!  ndimx               i4  i   x-dimension of mfeld
!  ndimy               i4  i   y-dimension of mfeld
!  nsize               i4  i   1/2 square size of region of interest
!  nx                  i4  i   x-value of center of region of interest
!  ny                  i4  i   y-value of center of region of interest
!  xms                 r4    o x-coordinate of cms
!  yms                 r4    o y-coordinate of cms


    integer     ndimx,ndimy
    real        ifeld(ndimx,ndimy)
    integer     i,j,nsize,nx,ny
    real        sum,xms,yms,xerr,yerr,itmp

! at the beginning sum,xms,yms has the value 0.0

    sum = 0.
    xms = 0.
    yms = 0.

    do j = ny-nsize,ny+nsize
      do i = nx-nsize,nx+nsize

! now sum sum,xms,yms
        itmp = ifeld(i,j)
        sum = sum + itmp
        xms = xms + i*itmp
        yms = yms + j*itmp

      enddo
    enddo

    xms = xms/sum
    yms = yms/sum

    xerr = 0.
    yerr = 0.

    do j = ny-nsize,ny+nsize
      do i = nx-nsize,nx+nsize

        itmp = ifeld(i,j)
        xerr = xerr + itmp*(i-xms)**2
        yerr = yerr + itmp*(j-yms)**2

      end do
    end do
   
    xerr = sqrt(xerr/(sum*(sum-1.)))
    yerr = sqrt(yerr/(sum*(sum-1.)))

    if (xerr .eq. 0.0) xerr = 0.5
    if (yerr .eq. 0.0) yerr = 0.5 

    xms = xms - 0.5
    yms = yms - 0.5

   end subroutine cmass


  !--------------------------------------------------------------------
  subroutine match (lfeld,ndimx,ndimy,nfeldx,nfeldy,lareax,lareay, &
                    ncom,kmerk,kold,xms,yms,xerr,yerr,iwind,       &
                    tmplike,likemin,    ierr)
  !--------------------------------------------------------------------

! indicates source-regions and calculates cmass
!  author : g.hasinger date:  7-mar-1985
!  update : gih        date: 26-jan-1987 14:47
!  update : cri        date: 22-jan-1988 12:20
!  update : ptk        date: 16-feb-1990 17:04
!  update : exsas      date: 03-aug-1993 17:03

!  the routine takes array lfeld which contains the flagged regions of
!  interest and identifies sources as a number (minimal 1) of contiguos
!  pixels. a source is regarded as 'new' if its highest pixel is at
!  least match pixels away from the center of mass of any other source.
!  the maximum value of lfeld within each island is handed to isnr.
!  the center of mass of each source region is calculated

!  isnr                r4    o array with max. snr for each sour. regi.
!  iwind               i4    o size of detection window for each source
!  kmerk               i4  i/o actual/new number of sources yet detec.
!  lareax              i4  i   number of 'dead' x-pixels
!                              at the edges of ifeld
!  lareay              i4  i   number of 'dead' y-pixels at the ed. ...
!  lfeld               r4  i/? field with snr-flags
!  ncom                i4  i   compression of current step
!  ndimx               i4  ?/? x-dimension of all square fields
!  ndimy               i4  ?/? y-dimension of all square fields
!  nfeldx              i4  i   x-length of field to search
!  nfeldy              i4  i   y-length of field to search
!  xms                 r4  i/o array with x-coord. of detected sources
!  yms                 r4  i/o array with y-coord. of detected sources
!  cmass               sr      calculates center of mass
!!! IT 2015/03 - detection of faint extended sources:
!  tmplike             r4  i   preliminary likelihoods, calculated by msrbkg
!  likemin             i4  i   parameter likemin (also: dm_snrm, snrm, snrmin)

   integer     ndimx,ndimy,   nn
   real, dimension(:,:) :: lfeld
   integer     i,l,ierr,iwind(max_sou),j,k,kmerk,kold,kold1, &
               lareax,lareay,ncom,nfeldx,nfeldy,nfound,nlx,nly, &
               nx,ny,ii,jj,mboxsize, xmin, xmax, ymin, ymax
   real        max,rat,dmatch,xm,xms(max_sou),ym,yms(max_sou)
   real        xe,xerr(max_sou),ye,yerr(max_sou),snr(max_sou)
   real        tmplike(max_sou)   !!! IT 2015/03
   real        likemin            !!! IT 2015/03

!!!!! declarations for test output image:
!      real(kind=single), dimension(:,:), pointer :: r4arr
!      integer, dimension(2) :: ldim  
!      type(DataSetT) :: set
!      Type(ArrayT) :: arr
!      integer :: ix,iy
!!!!!!!!!!!!!!!!!!!!!!!!!

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  
! write lfeld to fits file:
!          ldim(1)= ndimx
!          ldim(2)=ndimy
!          set = dataset("lfeld0.fits", write)
!          arr = addArray(set, "PRIMARY", Real32, ldim)
!          r4arr => real32array2data(arr)
!          do ix=1,ldim(1)
!            do iy=1,ldim(2)
!              r4arr(ix-1,iy-1)=lfeld(ix,iy)
!            end do
!          end do 
!          call release(arr)
!          call release(set)   
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!          

   ierr=0

   nlx = nfeldx-lareax+1
   nly = nfeldy-lareay+1
!!!   j = lareay
!!!   i = lareax

   !!! largest possible subimage area in image coordinates
   xmin = minval(ixb0(1:nim))/ncom+1
   ymin = minval(iyb0(1:nim))/ncom+1
   xmax = maxval(ixb1(1:nim))/ncom+1
   ymax = maxval(iyb1(1:nim))/ncom+1

! Always use 3x3 box here (02/11/04 GL):
   mboxsize=1

                              !!! restrict to subimage area in image coordinates
!!!   do while (j.le.nly .and. ierr.le.1)
!!!     if (i.gt.nlx) i = lareax
!!!     do while (i.le.nlx .and. ierr.le.1)
    do j=ymin,ymax
      do i=xmin,xmax

!      if source region : lfeld(i,j) > 0
       if (lfeld(i,j).gt.0) then
! find pixel of source region with maximum value
         nx = i
         ny = j
         max = lfeld(nx,ny)
         k = nx - mboxsize
         l = ny - mboxsize

         do while (l.le.ny+mboxsize)
           if (lfeld(k,l).gt.max) then
             nx = k
             ny = l
             k = nx - mboxsize
             l = ny - mboxsize
             max = lfeld(nx,ny)
           else
             k = k + 1
             if (k.gt.nx+mboxsize) then
               k = nx - mboxsize
               l = l + 1
             endif
           endif
         enddo


! compare with all sources already found before acceptance
! if the difference in one coordinate is less than dmatch
! pixels or source too close to edge of allowed region
! neglect the source
         nfound = 0
         if ( ((nx-lareax).lt.2 .or. (nlx-nx).lt.2 .or. &
               (ny-lareay).lt.2 .or. (nly-ny).lt.2) .and. &
                ncom.lt.8 ) nfound = 1

!        if (source found)
         if (nfound.eq.0) then

           if (kmerk.gt.max_sou-1) then
             kold1 = kold+1
             do ii = kold1, kmerk     
               do jj = 1, kmerk
                 if (ii /= jj .and. snr(ii) > 0.0 .and. snr(jj) > 0.0) then
                   rat = snr(ii)/snr(jj) 
                   if (rat < 1.0) rat = 1.0/rat
                   dmatch = (0.5+boxsize)*(iwind(ii)+iwind(jj))-1.0
                   if (abs(xms(ii)-xms(jj)) < dmatch .and. &
                       abs(yms(ii)-yms(jj)) < dmatch) then
!                     write(*,*) "match 1:",boxsize,iwind(ii),xms(ii),yms(ii),dmatch
                     if (snr(ii) < snr(jj) .or. jj < kold1) then
                       snr(ii) = -1.0
                     else
                       snr(jj) = -1.0
                     end if
                   end if
                 end if
               end do
             end do

             ii = kold1
             do while (ii <= kmerk)
               if (snr(ii) < 0.0) then
                 xms(ii) = xms(kmerk)
                 yms(ii) = yms(kmerk)
                 xerr(ii) = xerr(kmerk)
                 yerr(ii) = yerr(kmerk)
                 iwind(ii) = iwind(kmerk)
                 snr(ii) = snr(kmerk)
                 kmerk = kmerk-1
               else
                 ii = ii+1
               end if
             end do
           end if

           kmerk = kmerk + 1

           if (kmerk.ge.max_sou) then
              call error ("BufferFull", &
                   "Source buffer overflow") 
           else

! nx and ny are the coordinates of the highest pixel
! calculates center of mass
             call cmass (nx,ny,lfeld,ndimx,ndimy,2,xm,ym,xe,ye)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
             !test nsize=1:
!             call cmass (nx,ny,lfeld,ndimx,ndimy,1,xm,ym,xe,ye) 
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
             xms(kmerk) = xm*ncom
             yms(kmerk) = ym*ncom
             xerr(kmerk) = xe*ncom
             yerr(kmerk) = ye*ncom
             iwind(kmerk) = ncom
             snr(kmerk) = max/10.0

           endif
         endif
       endif
!!!       i = i + 1
     enddo
!!!     j = j + 1
   enddo

   kold1 = kold+1
   do i = kold1, kmerk     
     do j = 1, kmerk
       if (i /= j .and. snr(i) > 0.0 .and. snr(j) > 0.0) then
         rat = snr(i)/snr(j) 
         if (rat < 1.0) rat = 1.0/rat
!        dmatch = (4.0*boxsize + min(20.0,0.4*rat))**2
!        if ((xms(i)-xms(j))**2+(yms(i)-yms(j))**2 < dmatch) then
!         dmatch = 2*boxsize*ncom
         dmatch = (0.5+boxsize)*(iwind(i)+iwind(j))-1.0
         if (abs(xms(i)-xms(j)) < dmatch .and. &
             abs(yms(i)-yms(j)) < dmatch) then
!             write(*,*) "match 2:",boxsize,iwind(i),xms(i),yms(i),dmatch
           !!! 2015/03 IT - detection of faint extended sources:
           !!! added check of tmplike(j). Detections from previous runs can be
           !!! overwritten by new - potentially more likely - ones if their
           !!! equivalent likelihood is below likemin.
           if (snr(i) < snr(j) .or. (j < kold1 .and. tmplike(j) >= likemin)) then
             snr(i) = -1.0
           else
             snr(j) = -1.0
           end if
         end if
       end if
     end do
   end do

   i = kold1
   do while (i <= kmerk)
     if (snr(i) < 0.0) then
       xms(i) = xms(kmerk)
       yms(i) = yms(kmerk)
       xerr(i) = xerr(kmerk)
       yerr(i) = yerr(kmerk)
       iwind(i) = iwind(kmerk)
       snr(i) = snr(kmerk)
       kmerk = kmerk-1
     else
       i = i+1
     end if
   end do

!   do i=1,kmerk
!     write(*,*) "match:",xms(i),yms(i),snr(i)
!   end do

 end subroutine match

!--------------------------------------------------------------------
  subroutine getfrac(boxsize,ncom,alpha,beta,lloc)
!--------------------------------------------------------------------

! determine fraction of PSF in detection box
! in a grid of 15 offax angles, 6 azimuth angles (tan(phi)=0.0-1.0), 
! and up to max_ima energies

  use cal

  integer    ienergy,ioff,boxsize,iphi,ncom,nx,ny,ix,iy
  real       alpha(15,6,max_ima),beta(15,6,max_ima),tot,xfrac,yfrac
  real(kind=double)        phiarr(6),ePsf
  real(kind=double)        boxHWmm,pix2mm
  real(kind=double)        bgHWmm,off
  real(kind=single), dimension(:,:), pointer :: pima
  logical lloc

  type (PsfT) ppsf
  type (PsfWindowT) PsfWindow
  type (PsfValidityRangesT) vranges
  type(PsfBinSizeT)  bsiz

  data phiarr /0d0,0.197d0,0.381d0,0.54d0,0.675d0,0.785d0/   

    pix2mm  = ncom*abs(crdelt1)*7500.*3.14159265/180.0
    boxHWmm = (boxsize+0.5)*pix2mm
!    bgHWmm  = (boxsize+1.5)*pix2mm 
     bgHWmm  = (boxsize+2.0)*pix2mm


     call CAL_setState1(instrument=EPN,AccuracyLevel=ACCURACY_MEDIUM)

     vranges=CAL_psfValidityRanges()
     
     do ienergy=1,nim
      if (index(instrume(nim),"PN") > 0 .or. &
          index(instrume(nim),"pn") > 0) then 
        call CAL_setState1(instrument=EPN,AccuracyLevel=ACCURACY_MEDIUM)   
      elseif (index(instrume(nim),"MOS1") > 0 .or. &
             index(instrume(nim),"mos1") > 0) then 
        call CAL_setState1(instrument=EMOS1,AccuracyLevel=ACCURACY_MEDIUM) 
      else 
        call CAL_setState1(instrument=EMOS2,AccuracyLevel=ACCURACY_MEDIUM) 
      endif
      do ioff=1,15
        do iphi=1,6
         off=(ioff-0.5)*60.0
         ePsf=max(vranges%eMin+1.,energy(ienergy)) 
         ePsf=min(vranges%eMax+1.,ePsf)
         off=max(vranges%thetaMin+1.,off) 
         off=min(vranges%thetaMax+1.,off)
         ppsf = Cal_getPSF(ePsf,off,phiarr(iphi))
         bsiz = CAL_psfBINSIZE(ppsf) 
         
         PsfWindow%halfX=boxHWmm
         PsfWindow%halfY=boxHWmm
         call CAL_psfImage (ppsf, PsfWindow, pima)
         nx = size(pima,1)
         ny = size(pima,1)
         xfrac=(nx*bsiz%binX/2.-PsfWindow%halfX)/bsiz%binX
         yfrac=(ny*bsiz%binY/2.-PsfWindow%halfY)/bsiz%binY

         if   (xfrac <= 0) then
           xfrac = abs(xfrac)+1.0
         else
           xfrac = 1.0 - xfrac
         endif
         if   (yfrac <= 0) then
           yfrac = abs(yfrac)+1.0
         else
           yfrac = 1.0 - yfrac  
         end if

         tot=0.0
         do ix=1,nx-2
           do iy=1,ny-2
             tot=tot+pima(ix,iy) 
           end do 
         end do
         do ix=0,nx-1
           tot=tot+pima(ix,0)*yfrac
           tot=tot+pima(ix,ny-1)*yfrac
         end do
         do iy=1,ny-2
           tot=tot+pima(0,iy)*xfrac
           tot=tot+pima(nx-1,iy)*xfrac
         end do             
         alpha(ioff,iphi,ienergy)=tot*bsiz%binX*bsiz%binY

         if (lloc) then
           PsfWindow%halfX=bgHWmm
           PsfWindow%halfY=bgHWmm
           call CAL_psfImage (ppsf, PsfWindow, pima)
           nx = size(pima,1)
           ny = size(pima,1)
           xfrac=(nx*bsiz%binX/2.-PsfWindow%halfX)/bsiz%binX
           yfrac=(ny*bsiz%binY/2.-PsfWindow%halfY)/bsiz%binY

           if   (xfrac <= 0) then
             xfrac = abs(xfrac)+1.0
           else
             xfrac = 1.0 - xfrac
           end if
           if   (yfrac <= 0) then
             yfrac = abs(yfrac)+1.0
           else
             yfrac = 1.0 - yfrac  
           endif
           tot=0.0
           do ix=1,nx-2
             do iy=1,ny-2
               tot=tot+pima(ix,iy) 
             end do 
           end do
           do ix=0,nx-1
             tot=tot+pima(ix,0)*yfrac
             tot=tot+pima(ix,ny-1)*yfrac
           end do
           do iy=1,ny-2
             tot=tot+pima(0,iy)*xfrac
             tot=tot+pima(nx-1,iy)*xfrac
           end do         

           beta(ioff,iphi,ienergy)=tot*bsiz%binX*bsiz%binY &
                                   -alpha(ioff,iphi,ienergy)
         end if
        end do
       end do
     end do


 end subroutine getfrac

  !--------------------------------------------------------------------
  subroutine mdetec (ndimx,ndimy,snrmin,nrun,ipa_ar,lareax, &
                    lareay,lmap,lloc,bkscal, &
                    kmerk,xms,yms,xerr,yerr,isnr,iwind,src,esrc, &
                    bkg,expos,rat,erat,raw_cts,raw_bkg,eef,maskzero,ierr)
  !--------------------------------------------------------------------

! source detection algorithm with background map
!  author : g.hasinger date: 22-may-1985
!  update : g.hartner  date: 10-jul-1985
!  update : grh        date: 24-mar-1986 11:56 correct exposure
!  update : cri        date: 22-jan-1988 12:24
!  update : ptk        date: 19-sep-1988 11:13
!  update : ptk        date: 17-nov-1992 21:07 absrib -> absri1
!  update : exsas      date: 03-aug-1993 17:12

!  sliding window source detection algorithm with background map
!  a 3x3 source window is slided over the field ifein, corresponding
!  background is taken from the background map kfein, signal/noise
!  ratio is estimated. whenever snr above the value snrmin a flag is
!  set in lfeld. the isolated regions of interest in lfeld are further
!  examined for consistency and center of mass. the mechanism is
!  repeated a number of times with increasing  contraction.

!  bkg                 r4  ?/? preliminary background per pixel
!  ifein               i2  i   input field (will not be changed)
!  ifeld               i2    o work buffer for input field
!  ipa_ar              i4      dimension of lareax and lareay
!  isnr                r4  ?/? maximum snr for each source region
!  iwind               i4  ?/? size of detection window for each source
!  kfein               r4  i   background map with the same pixel size
!                              as ifein
!  kfeld               r4    o work buffer for background map
!  kmerk               i4  i/? number of detected sources
!  lareax              i4  ?/? number of 'dead' x-rows in each step
!  lareay              i4  ?/? number of 'dead' y-columns in each step
!  lfeld               r4  ?/? field with snr-flag
!  lmap                l4  i/? flag for use of an exposure map (t=yes)
!  mfein               l1  i/? exposure map with the same pixel size
!                              as ifein (logical*4 !)
!  mfeld               l1    o work buffer for exposure map
!  ndimx               i4  ?/? x-dimension of all square fields
!  ndimy               i4  ?/? y-dimension of all square fields
!  nfeinx              i4  i/? x-length of field to search
!  nfeiny              i4  i/? y-length of field to search
!  nrun                i4  i   number of compress steps to be performed
!  snrmin              r4      signal-to-noise ratio threshold
!  src                 r4  ?/? preliminary source counts
!  xms                 r4  ?/? array, x-coordinates of detected sources
!  yms                 r4  ?/? array, y-coordinates of detected sources

!  absri1              sr      gives distance to the next rib or edge
!  abssrc              sr      gives distance to the next source
!  hflag               r4      output flag handling routine
!  match               sr      isolate regions of interest within lfeld
!  msnrmp              sr      calculates the map of snr's: lfeld
!  msrbkg              sr      gives preliminary src and bkg cts
!  rclnr               sr      cleans an array
!  rcmpi2              sr      compresses array by a factor of 2
!  compl4              sr      compresses array by a factor of 2

   integer     ipa_ar,ndimx,ndimy

   integer(kind=int16), dimension(:,:,:), allocatable :: ifeld
   logical, dimension(:,:,:), allocatable :: mfeld
   real, dimension(:,:,:), allocatable :: efeld, kfeld, kfeld_err, bfeld
   real, dimension(:,:), allocatable :: lfeld 

   integer    i,ierr,irun,istat,iwind(max_sou),j, &
              kmerk,kold,lareax(ipa_ar),lareay(ipa_ar),ncom, &
              nfeldx,nfeldy,nfelkx,nfelky,nfelmx, &
              nfelmy,nfelex,nfeley,nrun,&
              ipoint,ii,jj,nsum
   integer    ixbmin, ixbmax, iybmin, iybmax, ixshift, jyshift, & !!! IT 2016/11
              xmin, xmax, ymin, ymax
!   integer     ixd3(16),iyd3(16),ixd5(24),iyd5(24),ixx,iyy 

   logical     lmap,lloc
   real        bkscal,bsum,snrmin
   real, dimension(max_sou) :: xms,yms,xerr,yerr
   !!! IT 2015/03 - detection of faint extended sources: preliminary likelihoods
   real, dimension(max_sou) :: tmplike
   real, dimension(:,:) :: src,esrc,rat,erat,bkg,isnr,expos, &
               raw_cts,raw_bkg,eef
   integer, dimension(:,:) :: maskzero
   real        alpha(15,6,max_ima),beta(15,6,max_ima)   
    
   ierr = 0

! initialize kold if new list
   if(kmerk.eq.0)kold=0

   !!! 2015/03 IT: initialize tmplike with likemin
   tmplike(:) = snrmin

! give the arrays ifein,mfein and kfein to the work arrays ifeld,
! mfeld and kfeld
   allocate (ifeld(nim, ixsize, iysize))    !!! subimage coordinates
   allocate (efeld(nim, ixsize, iysize))
   allocate (kfeld(nim, ixsize, iysize))
   allocate (kfeld_err(nim, ixsize, iysize))
   allocate (mfeld(nim, ixsize, iysize))
   allocate (lfeld(npix(1), npix(2)))       !!! image coordinates

   nfeldx = npix(1)
   nfeldy = npix(2)

   call message(AppMsg,Verbose,"Performing source detection ...")

   do ipoint=1,nim
      ifeld(ipoint, 1:xbsize(ipoint), 1:ybsize(ipoint)) = &
           image(ipoint, 1:xbsize(ipoint), 1:ybsize(ipoint))

      efeld(ipoint, 1:xbsize(ipoint), 1:ybsize(ipoint)) = &
           exp_image(ipoint, 1:xbsize(ipoint), 1:ybsize(ipoint))
   end do

   if (.not. lloc) then 
     allocate (bfeld(nim, ixsize, iysize))
     do ipoint=1,nim
        bfeld(ipoint, 1:xbsize(ipoint), 1:ybsize(ipoint)) = &
             bg_image(ipoint, 1:xbsize(ipoint), 1:ybsize(ipoint))
     end do
   endif

   if (lmap) then
      do ipoint=1,nim
         mfeld(ipoint, 1:xbsize(ipoint), 1:ybsize(ipoint)) = &
              mask_image(ipoint, 1:xbsize(ipoint), 1:ybsize(ipoint))
      end do
   end if

! loop over nrun steps of compression 2
   ncom = 1
   irun = 1
   do while (irun.le.nrun .and. ierr.le.1)

   if (irun.gt.1) ncom = 2*ncom

   !!! largest possible subimage area in image coordinates
   xmin = minval(ixb0(1:nim))/ncom+1
   ymin = minval(iyb0(1:nim))/ncom+1
   xmax = maxval(ixb1(1:nim))/ncom+1
   ymax = maxval(iyb1(1:nim))/ncom+1

   if (lloc) then
!!!   do j = 1,nfeldy    !!! restrict to subimage area in image coordinates
!!!     do i = 1,nfeldx
   do j = ymin,ymax
     do i = xmin,xmax
       do ipoint = 1,nim
           bsum=0.
           nsum=0 
          !! IT 2016/11 continue, if pixels are within the subimage
          ixbmin = ixb0(ipoint)/ncom+1
          ixbmax = ixb1(ipoint)/ncom
          iybmin = iyb0(ipoint)/ncom+1
          iybmax = iyb1(ipoint)/ncom
          if ((i.ge.ixbmin).and.(i.le.ixbmax).and. &
              (j.ge.iybmin).and.(j.le.iybmax)) then

           do ii=i-boxsize-2,i+boxsize+2 
             if (ii.ge.ixbmin .and. ii.le.ixbmax) then
             ixshift = ii-ixbmin+1
             do jj=j-boxsize-2,j-boxsize-1
               if (jj.ge.iybmin .and. jj.le.iybmax) then  
               jyshift = jj-iybmin+1
               if (.not.lmap .or. mfeld(ipoint, ixshift, jyshift)) then  
                 bsum=bsum+ifeld(ipoint, ixshift, jyshift)
                 nsum=nsum+1
               end if 
               endif
             end do
             do jj=j+boxsize+1,j+boxsize+2
               if (jj.ge.iybmin .and. jj.le.iybmax) then 
               jyshift = jj-iybmin+1
               if (.not.lmap .or. mfeld(ipoint, ixshift, jyshift)) then  
                 bsum=bsum+ifeld(ipoint, ixshift, jyshift)
                 nsum=nsum+1
               end if  
               end if
             end do
             endif
           end do
           do jj=j-boxsize,j+boxsize 
             if (jj.ge.iybmin .and. jj.le.iybmax) then  
             jyshift = jj-iybmin+1
             do ii=i-boxsize-2,i-boxsize-1
               if (ii.ge.ixbmin .and. ii.le.ixbmax) then 
               ixshift = ii-ixbmin+1
               if (.not.lmap .or. mfeld(ipoint, ixshift, jyshift)) then  
                 bsum=bsum+ifeld(ipoint, ixshift, jyshift)
                 nsum=nsum+1
               end if
               end if
             end do
             do ii=i+boxsize+1,i+boxsize+2
              if (ii.ge.ixbmin .and. ii.le.ixbmax) then 
              ixshift = ii-ixbmin+1
              if (.not.lmap .or. mfeld(ipoint, ixshift, jyshift)) then  
                 bsum=bsum+ifeld(ipoint, ixshift, jyshift)
                 nsum=nsum+1
               end if    
               end if          
             end do
             endif
           end do

           ixshift = i-ixbmin+1
           jyshift = j-iybmin+1
           if (nsum > 0) then
             kfeld(ipoint, ixshift, jyshift) = bsum/float(nsum)
             kfeld_err(ipoint, ixshift, jyshift) = sqrt(bsum)/float(nsum)
           else
             kfeld(ipoint, ixshift, jyshift) = 0.0
           endif

           endif !!! coordinate check

       enddo
     enddo
   enddo

   else  !!! map mode

!!!   do j = 1,nfeldy    !!! restrict to subimage area in image coordinates
!!!     do i = 1,nfeldx
   do j = ymin,ymax
     do i = xmin,xmax
       do ipoint = 1,nim

           !average background map over detection box:
           bsum=0
           nsum=0
          !! IT 2016/11 continue, if pixels are within the subimage
          ixbmin = ixb0(ipoint)/ncom+1
          ixbmax = ixb1(ipoint)/ncom
          iybmin = iyb0(ipoint)/ncom+1
          iybmax = iyb1(ipoint)/ncom
          if ((i.ge.ixbmin).and.(i.le.ixbmax).and. &
              (j.ge.iybmin).and.(j.le.iybmax)) then

           do jj=j-boxsize,j+boxsize
            if (jj.ge.iybmin .and. jj.le.iybmax) then 
            jyshift = jj-iybmin+1
            do ii=i-boxsize,i+boxsize
              if (ii.ge.ixbmin .and. ii.le.ixbmax) then
                 ixshift = ii-ixbmin+1
                 if (.not.lmap .or. mfeld(ipoint, ixshift, jyshift)) then 
                   bsum=bsum+bfeld(ipoint, ixshift, jyshift)
                   nsum=nsum+1
                 endif
               endif
             enddo
             endif
           end do
           ixshift = i-ixbmin+1
           jyshift = j-iybmin+1
           if (nsum > 0) then
             kfeld(ipoint, ixshift, jyshift) = bsum/float(nsum)
           else
             kfeld(ipoint, ixshift, jyshift) = 0.0
           endif 
!------------------------------------------------
!          kfeld(ipoint,i,j) = bfeld(ipoint,i,j)
!------------------------------------------------
          end if  !!! coordinate check
       end do 
     enddo
   enddo

   endif  !!! local or map mode

! determine fraction of PSF in detection box:
   if (indetcoords) then    !!! IT 2015/06 no PSF correction for input images
                            !!!            in detector coordinates
      alpha(:,:,:) = 1.
      beta(:,:,:) = 0.
   else
      if (ncom == 1)   call getfrac(boxsize,ncom,alpha,beta,lloc) 
   end if

     write(errstr,"(a,i3)") "Compression step",ncom  
     call message(AppMsg,Verbose)

! clean lfeld
     lfeld = 0.

! calculation of signal-to-noise map
       call msnrmp (ifeld,kfeld,mfeld,ndimx,ndimy,nfeldx,nfeldy, &
                  snrmin,lareax(irun),lareay(irun),lmap,bkscal,lfeld, &
                  ncom,istat)
! now perform match and cms
       call match (lfeld,ndimx,ndimy,nfeldx,nfeldy,lareax(irun), &
                   lareay(irun),ncom,kmerk,kold,xms,yms,xerr,yerr, &
                   iwind,  tmplike,snrmin,  istat)   !!! "snrmin" is "likemin" 

! calculate preliminary source and background counts
       call msrbkg (ifeld,kfeld,kfeld_err,mfeld,efeld, &
                    ndimx,ndimy,ncom,lmap,lloc,kold,kmerk,xms,yms, &
                    src,esrc,bkg,expos,rat,erat,isnr,alpha,beta, &
                    raw_cts,raw_bkg,eef,maskzero,   tmplike,   istat)


! compress the arrays ifeld, mfeld, kfeld, and efeld by a factor 2
            nfelmx = nfeldx
            nfelmy = nfeldy
            nfelkx = nfeldx
            nfelky = nfeldy
            nfelex = nfeldx
            nfeley = nfeldy

            call rcmpi2 (nfeldx,nfeldy,ifeld,2,ncom)
            if (.not. lloc) call rcmpr4 (nfelkx,nfelky,bfeld,2,ncom)
            call rcmpr4 (nfelex,nfeley,efeld,2,ncom)

            if (lmap) call compl4(nfelmx,nfelmy,mfeld,2,ncom,0.55)

      irun = irun + 1
   enddo

  end subroutine mdetec

  !--------------------------------------------------------------------
  subroutine msnrmp (ifeld,kfeld,mfeld,ndimx,ndimy,nfeldx, &
                     nfeldy,snrmin,lareax,lareay,lmap,bkscal,lfeld,ncom,ierr)
  !--------------------------------------------------------------------

! calculates signal-to-noise map
    integer     ndimx,ndimy,ibox,jbox,ncom
    integer(kind=int16), dimension(:,:,:)  :: ifeld
    real, dimension(:,:,:)  :: kfeld
    real, dimension(:,:)  :: lfeld
    logical, dimension(:,:,:) :: mfeld

    integer   iso,mso,ipoint,nzero 
    integer   i,ierr,ii,istat,j,jj,lareax,lareay, &
              mlevel,nfeldx,nfeldy,nlx,nly
    integer   ixbmin, ixbmax, iybmin, iybmax, ixshift, jyshift, & !!! IT 2016/11
              xmin, xmax, ymin, ymax
    real      bkg,ibk,bkscal,snr,snrmin,lsum,smso
    !!! IT 2015/03: introduce new variable to allow for different values of
    !!! the minimum background, depending on the observation mode (pointing /
    !!! slew). It was 0.2 originally and changed to 0.1 on 2004/12/01.
    real       bkmin
    logical    lmap
    real, dimension(5,5) :: kernel 
    real :: ksum
 

!     data                       kernel/0.2,0.3,0.4,0.3,0.2, &
!                                       0.3,0.5,0.6,0.5,0.3, &
!                                       0.4,0.7,1.0,0.7,0.4, &
!                                       0.3,0.5,0.6,0.5,0.3, &
!                                       0.2,0.3,0.4,0.3,0.2/


     data                       kernel/0.2,0.3,0.4,0.3,0.2, &
                                       0.3,0.7,0.8,0.7,0.3, &
                                       0.4,0.8,1.0,0.8,0.4, &
                                       0.3,0.7,0.8,0.7,0.3, &
                                       0.2,0.3,0.4,0.3,0.2/



!     data                       kernel/0.4,0.5,0.6,0.5,0.4, &
!                                       0.5,0.7,0.8,0.7,0.5, &
!                                       0.6,0.8,1.0,0.8,0.6, &
!                                       0.5,0.7,0.8,0.7,0.5, &
!                                       0.4,0.5,0.6,0.5,0.4/



    
          !    test with flat kernel:
    !!! 2015/03 IT/RDS:
    !!! use flat kernel and higher minimum background for slew-mode data
    if (obsmode == "pointing") then   ! default
       kernel=kernel*2.5
       bkmin=0.1
    else                              ! slew data
       kernel(:,:)=1.0
       bkmin=0.2
    end if

    if (boxsize == 1) then
      mlevel=5
    else
      mlevel=15
    end if
 
! set values nlx, nly and start values for following loops
    ierr = 0
!!!    nlx = nfeldx - lareax + 1
!!!    nly = nfeldy - lareay + 1
!!!    i = lareax
!!!    j = lareay
    lsum = 0.

    lfeld = 0.

    !!! largest possible subimage area in image coordinates
    xmin = minval(ixb0(1:nim))/ncom+1
    ymin = minval(iyb0(1:nim))/ncom+1
    xmax = maxval(ixb1(1:nim))/ncom+1
    ymax = maxval(iyb1(1:nim))/ncom+1

                              !!! restrict to subimage area in image coordinates
!!!    do while (j.le.nly .and. ierr.le.1)
!!!      if (i.gt.nlx) i = lareax
!!!      do while (i.le.nlx .and. ierr.le.1)
    do j=ymin,ymax
      do i=xmin,xmax 

! source pixels from image ifeld
        snr = 0.
        nzero=0 
        do ipoint = 1,nim 

          !! IT 2016/11 continue, if pixels are within the subimage
          ixbmin = ixb0(ipoint)/ncom+1
          ixbmax = ixb1(ipoint)/ncom
          iybmin = iyb0(ipoint)/ncom+1
          iybmax = iyb1(ipoint)/ncom
          if ((i.ge.ixbmin).and.(i.le.ixbmax).and. &
              (j.ge.iybmin).and.(j.le.iybmax)) then

!!!       iso = 0
          smso = 0.0


          ibk = 0.
          mso = 0

!!!          ibox=0
!!!          jbox=0
          ksum=0.0
          do jj = j-boxsize,j+boxsize
            if (jj.ge.iybmin .and. jj.le.iybmax) then
            jyshift = jj-iybmin+1
            jbox = jj-j+boxsize+1  !! jbox+1
!!!            ibox=0
            do ii = i-boxsize,i+boxsize
              if (ii.ge.ixbmin .and. ii.le.ixbmax) then
              ixshift = ii-ixbmin+1
              ibox = ii-i+boxsize+1 !! ibox+1
                if (.not.lmap .or. mfeld(ipoint, ixshift, jyshift) &
                             .and. mfeld(ipoint, i-ixbmin+1, j-iybmin+1)) then
                  if (boxsize == 2) then

!                    iso = iso + ifeld(ipoint,ii,jj)*kernel(ibox,jbox)
                     smso = smso +  real(ifeld(ipoint, ixshift, jyshift))*kernel(ibox,jbox)

                    ksum = ksum +  kernel(ibox,jbox)  
                  else

!                    iso = iso + ifeld(ipoint,ii,jj)
                    smso = smso + real(ifeld(ipoint,ixshift, jyshift))
                    ksum=ksum+1
                  endif                        
!                  ibk = ibk + kfeld(ipoint,ii,jj)
                  mso = mso + 1
                endif
              end if 
            enddo
            endif
          enddo
          ixshift = i-ixbmin+1
          jyshift = j-iybmin+1
          if (.not.lmap .or. mfeld(ipoint, ixshift, jyshift)) then
!          ibk = max(mso*kfeld(ipoint,i,j),0.2)
!TEST:
!           changed minimum value ob ibk to 0.1 (from 0.2) 1/12/04
             !!! IT 2015/03: use variable bkmin, depending on obsmode
           ibk = max(ksum*kfeld(ipoint, ixshift, jyshift), bkmin)

!          ibk = mso*kfeld(ipoint,i,j)
          endif                   

! if iso or mso smaller than critical don't continue
           bkg = ibk*bkscal

!          if (lloc) then
!             if (iso == 0) then
!               snr = 0.0
!             else 
!               if (kfeld(ipoint,i,j) == 0 .or. kfeld_err(ipoint,i,j) == 0) then
!                 nback = 0
!               else 
!                 nback = nint( (kfeld(ipoint,i,j) / kfeld_err(ipoint,i,j) )**2)
!               endif
!               fak=1.0
!               if (nback <= 20) then
!                 fak=1.0
!                 do ifak=1,nback
!                   fak=fak*ifak 
!                 end do
!               else
!                 sig=sqrt(float(nback))
!               endif
!
!               bmin=max(1,nback-10)
!               bmax=nback+10
!               like=0.0 
!               norm=0.0
!               do kback=bmin,bmax
!                 back=kback-0.5
!                 if (nback == 0) then
!                   bk=back
!                 else
!                   bk=mso*back/float(nback)*kfeld(ipoint,i,j)
!                 endif
!                 if (nback <= 20) then   
!                   p0 = (back**nback)/fak*exp(-back)
!                   like = like+p0*exp(-poisl2(float(iso),bk,istat))
!                   norm = norm+p0
!                 else
!                   p0 = exp(-0.5*(back-nback)**2/sig**2)
!                   like = like+p0*exp(-poisl2(float(iso),bk,istat))
!                   norm = norm+p0
!                 endif
!               end do 
!               if (like > 0) then
!                 snr = snr-dlog(like/norm) 
!               else 
!                 bk = mso*kfeld(ipoint,i,j) 
!                 snr = snr + poisl2(float(iso),bk,istat)
!               endif 
!             endif  
!          else 

!            if (iso > 0 .and. mso > mlevel) &
             if (smso > 0.0 .and. mso > mlevel) &

!               snr = snr + poisl2 (float(iso),bkg,istat)
                snr = snr + poisl2 (smso,bkg,istat)

            if  (mso .le. mlevel) nzero=nzero+1 
               

!          endif

            endif !!! coordinate check

        end do

        if (nim > nzero) snr = gammq(real(nim-nzero),snr) 

!        if (snr.ge.snrmin) then
! test lower likelihood level for map
        if (snr.ge.snrmin-1.0) then
          snr = snr*10.
          lfeld(i,j) = snr
          lsum = lsum + snr
        endif

!!!        i = i + 1
      enddo
!!!      j = j + 1
    enddo


  end subroutine msnrmp

  !--------------------------------------------------------------------
   subroutine msrbkg (ifeld,kfeld,kfeld_err,mfeld,efeld, &
                      ndimx,ndimy,ncom,lmap,lloc,kold,kmerk,xms,yms, &
                      src,esrc,bkg,expos,rat,erat,isnr,alpha,beta, &
                      raw_cts,raw_bkg,eef,maskzero,   tmplike,    ierr)
  !--------------------------------------------------------------------

! calculates source and background fluxes for mdetect

! author : g.hasinger date: 25-may-1985
! update : gih        date: 4-aug-1987 18:39
!  update : cri        date: 22-jan-1988 12:23
!  update : ptk        date: 16-nov-1990 08:19
!  update : exsas               date: 03-aug-1993 17:12

!  takes the sources detected by mdetec and calculates src and bkg
!  using the image ifeld and the background map kfeld

!  bkg                 r4    o background counts per pixel
!  ifeld               i2  i   input field
!  isnr                r4    o signal to noise ratio
!  kfeld               r4  i   background map
!  kmerk               i4  i   number of newly detected sources
!  kold                i4  i/o number of old sources in list (updated)
!  lmap                l4  i   flag for use of an exposure map (t=yes)
!  mfeld               l1  i   exposure map with same size as ifeld
!                              (logical*4 !)
!  ncom                i4  i   compression of current step
!  ndimx               i4      x-dim of all square fields (in memory)
!  ndimy               i4      y-dim of all square fields (in memory)
!  src                 r4    o source counts
!  xms                 r4  i   center of mass x-coordinate from match
!  yms                 r4  i   center of mass y-coordinate from match

!  poisl2              double      calculates signal to noise ratio

!!! IT 2015/03 - detection of faint extended sources:
!  tmplike             r4    o temporary detection likelihoods

   integer(kind=int16), dimension(:,:,:) :: ifeld 
   real, dimension(:,:,:) :: kfeld, kfeld_err,efeld
   logical, dimension(:,:,:) :: mfeld 
   integer ::  i,ierr,ii,istat,j,jj,kmerk,kold,n, &
               ncom,ipoint,ioff,ndimx,ndimy,iphi,    nzero !!! IT 2015/03: nzero
   !!!(IT 03/2015) iii,jjj,iii0,jjj0,     ! dummy loops - no longer needed
   integer     ixbmin, ixbmax, iybmin, iybmax, ixshift, jyshift  !!! IT 2016/11
   integer, dimension(max_ima) :: mso,iso 
   real ::     aa,bb,kk,ke,alpha0,beta0,bk,bkmin, &
               bkgsum,isn,deltax,deltay
   real, dimension(max_ima) :: imax 
   real, dimension(15,6,max_ima) :: alpha,beta
   real, dimension(max_sou) :: xms,yms
   real, dimension(max_sou) :: tmplike        !!! IT 2015/03
   real, dimension(:,:) :: src,esrc,bkg,rat,erat,expos, &
               isnr,raw_cts,raw_bkg,eef
   integer, dimension(:,:) :: maskzero
   logical     lmap,lloc

!  real f(10)
!  integer i

!        character*8   rtname
!  data          alpha,beta /0.78,0.17/

! this is for the AIP simulation PSF:
!   data          alpha3,beta3 /0.40,0.22/
!   data          alpha5,beta5 /0.62,0.11/

    if (boxsize == 1) then
       bkmin=1./16.
    else
       bkmin=1./24.
    end if

! calculate bkg and src
!
! to calculate fluxes the amount of source flux within the source
! and background cells (alpha and beta resp.) have to be known

    ierr = 0
    mso = 9
    n = kold + 1
    bkgsum = 0.


    do while (n.le.kmerk .and. ierr.le.1)

! for test purposes try various locations 
     imax = -999.
!     do jjj = -1,1
!       do iii = -1,1
!!!(IT 03/2015)     do jjj = 0,0
!!!(IT 03/2015)       do iii = 0,0


! 3x3 or 5x5 source pixels from image ifeld
         i = nint(xms(n)/ncom) !!!(IT 03/2015)   + 0.5*iii
         j = nint(yms(n)/ncom) !!!(IT 03/2015)   + 0.5*jjj
!         i=max(1,i)
!         j=max(1,j) 
        

         mso = 0
         iso = 0

         !!! IT 2015/03: initialize nzero and preliminary likelihood of source n
         nzero = 0
         tmplike(n) = 0.

         do jj = j-boxsize,j+boxsize
           do ii = i-boxsize,i+boxsize
               do ipoint=1,nim
                 ixbmin = ixb0(ipoint)/ncom+1
                 ixbmax = ixb1(ipoint)/ncom
                 iybmin = iyb0(ipoint)/ncom+1
                 iybmax = iyb1(ipoint)/ncom
                 if ((ii.ge.ixbmin).and.(ii.le.ixbmax).and. &
                     (jj.ge.iybmin).and.(jj.le.iybmax)) then
                    ixshift = ii-ixbmin+1
                    jyshift = jj-iybmin+1

                 if (.not.lmap.or.mfeld(ipoint, ixshift, jyshift)) then
                   mso(ipoint) = mso(ipoint) + 1
                   iso(ipoint) = iso(ipoint) + ifeld(ipoint, ixshift, jyshift)
                 end if
                 end if   !!! end coordinate check
               end do
           end do
         end do
 
         do ipoint = 1,nim
          !! IT 2016/11 continue, if pixels are within the subimage
          ixbmin = ixb0(ipoint)/ncom+1
          ixbmax = ixb1(ipoint)/ncom
          iybmin = iyb0(ipoint)/ncom+1
          iybmax = iyb1(ipoint)/ncom
          if ((i.ge.ixbmin).and.(i.le.ixbmax).and. &
              (j.ge.iybmin).and.(j.le.iybmax)) then
           ixshift = i-ixbmin+1
           jyshift = j-iybmin+1

         ! get fraction of psf in detection box
           if (ncom == 1) then
             deltax=abs(xms(n)-xcenters(ipoint))
             deltay=abs(yms(n)-ycenters(ipoint))
             ioff=min(15,int(sqrt((deltax*crdelt1)**2+ &
                                  (deltay*crdelt2)**2)*60.0)+1)
!!! IT 2012/06/21: catch the case deltax=deltay=0, where iphi has to be 1
!!! (i.e. source at same image pixels (XCEN,YCEN) as optical axis)
! iphi is also =1 if only one of the distances is zero. So we can include
! those cases, too, and leave the "if" statements as soon as possible.
             if (deltax .eq. 0. .or. deltay .eq. 0.) then
                iphi = 1
             else
                if (deltax > deltay) then
                   iphi=int(5*deltay/deltax)+1
                else
                   iphi=int(5*deltax/deltay)+1
                end if
             end if
             alpha0=alpha(ioff,iphi,ipoint)
             beta0=beta(ioff,iphi,ipoint)
!             beta0=0
           else
             beta0 = 0.0
             alpha0 = 1.0
           end if
           if (lloc) then
             bk = mso(ipoint)*max(kfeld(ipoint, ixshift, jyshift),bkmin)
             isn =  poisl2(float(iso(ipoint)),bk,istat)
           else 
             bk = mso(ipoint)*kfeld(ipoint, ixshift, jyshift)
             isn =  poisl2(float(iso(ipoint)),bk,istat)
           endif
             if (mso(ipoint) .eq. 0) then
                maskzero(ipoint,n)=1
                !!! IT 2015/03: calculate nzero and sum of SNRs
                nzero = nzero+1
             else
                tmplike(n) = tmplike(n) + isn
             endif
             imax(ipoint) = isn
             isnr(ipoint,n) = isn
             eef(ipoint,n) = alpha0
             raw_cts(ipoint,n) = iso(ipoint) 
             raw_bkg(ipoint,n) = kfeld(ipoint, ixshift, jyshift)/(ncom*ncom) 
!!!(IT 03/2015)             iii0=iii
!!!(IT 03/2015)             jjj0=jjj
             if (lloc) then
!              bb = beta0/(8.0*(1+boxsize))
               bb = beta0/(16*boxsize+24)
               aa = alpha0-mso(ipoint)*bb
               kk = kfeld(ipoint, ixshift, jyshift)
               ke = kfeld_err(ipoint, ixshift, jyshift) 
               src(ipoint,n) = (iso(ipoint)-mso(ipoint)*kk)/aa
               esrc(ipoint,n) = sqrt(iso(ipoint)+(mso(ipoint)*ke)**2)/aa
               bkg(ipoint,n) = (kk-src(ipoint,n)*bb)/ncom**2
               if (bkg(ipoint,n) <= 0.0) then
                 src(ipoint,n)=iso(ipoint)/alpha0
                 esrc(ipoint,n)=sqrt(real(iso(ipoint)))/alpha0
                 bkg(ipoint,n)=0.0
               end if
             else
               src(ipoint,n)=(iso(ipoint)-bk) / alpha0
               esrc(ipoint,n)=sqrt(real(iso(ipoint)))/alpha0 
               bkg(ipoint,n) = bk / (mso(ipoint)*ncom*ncom)
             end if
             expos(ipoint,n) = efeld(ipoint, ixshift, jyshift)
             rat(ipoint,n) = src(ipoint,n)/expos(ipoint,n)
             erat(ipoint,n)= esrc(ipoint,n)/expos(ipoint,n) 
!           endif

         endif !!! end coordinate check

         enddo
!!!(IT 03/2015)       enddo
!!!(IT 03/2015)     enddo

         !!! IT 2015/03 - detection of faint extended sources:
         !!! calculate preliminary equivalent likelihood of source n
         tmplike(n) = gammq(real(nim-nzero), tmplike(n))

     n = n + 1
   enddo
   kold = kmerk

!   do i=1,kmerk
!     write(*,*) "msrbkg:",xms(i),yms(i),isnr(1:5,i)
!   end do

  end subroutine msrbkg 

  !--------------------------------------------------------------------
  function poisl2(measur,expect,ierr) result (result_poisl2)
  !--------------------------------------------------------------------

!  poiss. likelihood

!  author : s.doebereiner date: 11-sep-1996

!  get logarithm of the cumulative poisson probability (likelihood)  
!  to have 'measur' or more counts if 'expect' counts are expected. 

!  expect              r4  i/? expected value (background)
!  ierr                i4    o = 0 no error
!                              = error occured in ppois
!  measur              i4  i/? measured value
!  poisl2              r4      likelihood

!  ppoiss              r4      interpolates snr in table
!  gammp               r4      calculates snr=-ln(p(measur,expect))
!  wrflag              sr      writes to flag output stream
!  wrerr               sr      write error message

    real(kind=double) :: result_poisl2
    integer       ierr
    real          expect,measur

    ierr = 0

    result_poisl2 = 0.0 

    if (measur > 0) then
      result_poisl2 = gammp(measur,expect)  
    endif

  end function poisl2

  !--------------------------------------------------------------------
  subroutine rclnr(r4feld,ipa_x,ipa_y,nact_x,nact_y)
  !--------------------------------------------------------------------

!  author : grh        date:
!  update : cri        date: 23-nov-1988 09:30
!  update : ptk        date: 06-mar-1990 07:34
!  update : exsas              date: 03-aug-1993 16:26

!  reset the input r*4 field to 0

!  ipa_x               i4      x-dimension of the array
!  ipa_y               i4      y-dimension of the array
!  r4feld              r4    o array to be resetted
!  nact_x              i4  i   x-dimension to be resetted
!  nact_y              i4      y-dimension to be resetted

!  hflag               r4      output flag handling routine
!  wrflag              sr      writes to flag output stream

    integer     i,ipa_x,ipa_y,j,nact_x,nact_y
    real        r4feld(ipa_x,ipa_y)

    do i=1,nact_x
      do j=1,nact_y
        r4feld(i,j) = 0.
      end do
   end do

   end subroutine rclnr


  !---------------------------------------------------------------------
  subroutine compl4(nact_x,nact_y,l4feld,ncomp,ncom,thresh)
  !---------------------------------------------------------------------

! compression of ncomp

!  author : grh        date:
!  update : cri        date: 23-nov-1988 10:58
!  update : ptk        date: 17-nov-1992 12:30 l*4 data
!  update : exsas      date: 03-aug-1993 16:19

!  compress a logical field of a factor ncomp
!  take the mean value of ncomp*ncomp and digitalize the result
!  depending of the threshold (0.-1.)

!  ipa_x               i4      x-dimension of the array
!  ipa_y               i4      y-dimension of the array
!  l4feld              l4  i/o array to be resetted
!  nact_x              i4  i/o actual x-dimension to be compressed (on
!                              exit is compressed of a factor ncomp)
!  nact_y              i4  i/o actual y-dimension to be compressed (on
!                              exit is compressed of a factor ncomp)
!  ncomp               i4  i   compression factor
!  thresh              r4      digital threshold (between 0. and 1.)

    integer     i,j,k,l,merk,nact_x,nact_y,ncomp,ncom,ipoint !!!, &
!!!                nnew_x,nnew_y
    integer     ixbmin, ixbmax, iybmin, iybmax  !!! IT 2016/11
    logical, dimension(:,:,:) :: l4feld
    real        thresh,value

    nact_x = nact_x/ncomp
    nact_y = nact_y/ncomp
    value = ncomp*ncomp*thresh
    do ipoint = 1,nim
      !! IT 2016/11 subimage coordinates
      ixbmin = ixb0(ipoint)/ncom/ncomp
      ixbmax = ixb1(ipoint)/ncom/ncomp
      iybmin = iyb0(ipoint)/ncom/ncomp
      iybmax = iyb1(ipoint)/ncom/ncomp
      do j=iybmin+1,iybmax
        do i=ixbmin+1,ixbmax
          merk = 0
          do l=(j-1)*ncomp+1,j*ncomp
            do k=(i-1)*ncomp+1,i*ncomp
              if(l4feld(ipoint,k-ixbmin*ncomp,l-iybmin*ncomp))   merk=merk+1
            end do
          end do
          if(merk.gt.value)   then
            l4feld(ipoint,i-ixbmin,j-iybmin) = .true.
          else
            l4feld(ipoint,i-ixbmin,j-iybmin) = .false.
          end if
        end do
      end do
    end do
!!!    nact_x = nnew_x
!!!    nact_y = nnew_y

  end subroutine compl4

  !--------------------------------------------------------------------
   subroutine rcmpi2(nact_x,nact_y,i2feld,ncomp,ncom)
  !--------------------------------------------------------------------

! compression of ncomp

!  author : grh        date:
!  update : cri        date: 23-nov-1988 10:58
!  update : ptk        date: 06-mar-1990 13:20
!  update : exsas      date: 03-aug-1993 16:26

!  ipa_x               i4  i   x-dimension of the array
!  ipa_y               i4  i   y-dimension of the array
!  i2feld              i2  i   array to be resetted
!  nact_x              i4    o actual x-dimension to be compressed (on
!                              exit is compressed of a factor ncomp)
!  nact_y              i4    o actual y-dimension to be compressed (on
!                              exit is compressed of a factor ncomp)
!  ncomp               i4 i    compression factor

!  hflag               r4      output flag handling routine
!  wrflag              sr      writes to flag output stream

    integer     i,j,k,l,merk,nact_x,nact_y,ncomp,ncom,ipoint !!!, &
!!!                nnew_x,nnew_y
    integer     ixbmin, ixbmax, iybmin, iybmax  !!! IT 2016/11
    integer(kind=int16), dimension(:,:,:) :: i2feld

    nact_x = nact_x/ncomp
    nact_y = nact_y/ncomp
    do ipoint=1,nim
       !! IT 2016/11 subimage coordinates
       ixbmin = ixb0(ipoint)/ncom/ncomp
       ixbmax = ixb1(ipoint)/ncom/ncomp
       iybmin = iyb0(ipoint)/ncom/ncomp
       iybmax = iyb1(ipoint)/ncom/ncomp
       do j=iybmin+1,iybmax
          do i=ixbmin+1,ixbmax
          merk = 0
          do l=(j-1)*ncomp+1,j*ncomp
            do k=(i-1)*ncomp+1,i*ncomp
              merk = merk+i2feld(ipoint, k-ixbmin*ncomp, l-iybmin*ncomp)
            end do
          end do
!         if(merk.gt.32000)   merk=32000
          i2feld(ipoint, i-ixbmin, j-iybmin) = merk
        end do
      end do
    end do
!!!    nact_x = nnew_x
!!!    nact_y = nnew_y

  end subroutine rcmpi2

  !--------------------------------------------------------------------
  subroutine rcmpr4(nact_x,nact_y,r4feld,ncomp,ncom)
  !--------------------------------------------------------------------

! compression of ncomp

!  author : grh        date:
!  update : cri        date: 23-nov-1988 10:58
!  update : ptk        date: 06-mar-1990 13:20
!  update : ptk        date: 06-mar-1990 13:46
!  update : exsas              date: 03-aug-1993 16:26

!  compress a r4-field of a factor ncomp

!  ipa_x               i4  i   x-dimension of the array
!  ipa_y               i4  i   y-dimension of the array
!  r4feld              r4  i   array to be resetted
!  nact_x              i4    o actual x-dimension to be compressed (on
!                              exit is compressed of a factor ncomp)
!  nact_y              i4    o actual y-dimension to be compressed (on
!                              exit is compressed of a factor ncomp)
!  ncomp               i4 i    compression factor

!  hflag               r4      output flag handling routine
!  wrflag              sr      writes to flag output stream

    integer     i,j,k,l,nact_x,nact_y,ncomp,ncom,ipoint !!!, &
!!!                nnew_x,nnew_y
    integer     ixbmin, ixbmax, iybmin, iybmax  !!! IT 2016/11
    real, dimension(:,:,:) :: r4feld
    real        merk

    nact_x = nact_x/ncomp
    nact_y = nact_y/ncomp
    do ipoint = 1,nim
       !! IT 2016/11 subimage coordinates
       ixbmin = ixb0(ipoint)/ncom/ncomp
       ixbmax = ixb1(ipoint)/ncom/ncomp
       iybmin = iyb0(ipoint)/ncom/ncomp
       iybmax = iyb1(ipoint)/ncom/ncomp
       do j=iybmin+1,iybmax
          do i=ixbmin+1,ixbmax
          merk = 0.0
          do l=(j-1)*ncomp+1,j*ncomp
            do k=(i-1)*ncomp+1,i*ncomp
              merk = merk+r4feld(ipoint, k-ixbmin*ncomp, l-iybmin*ncomp)
            end do
          end do
          r4feld(ipoint, i-ixbmin, j-iybmin) = merk
        end do
      end do
    end do
!!!    nact_x = nnew_x
!!!    nact_y = nnew_y

  end subroutine rcmpr4

end module eboxdetect_mod





